/**
 * Copyright (c) 2007, Gaudenz Alder
 */
part of graph.model;


/**
 * Implements a mechanism for temporary cell Ids.
 */
class CellPath {

  /**
   * Defines the separator between the path components. Default is
   * <code>.</code>.
   */
  static String PATH_SEPARATOR = r"\\.";

  /**
   * Creates the cell path for the given cell. The cell path is a
   * concatenation of the indices of all cells on the (finite) path to
   * the root, eg. "0.0.0.1".
   * 
   * @param cell Cell whose path should be returned.
   * @return Returns the string that represents the path.
   */
  static String create(ICell cell) {
    String result = "";

    if (cell != null) {
      ICell parent = cell.getParent();

      while (parent != null) {
        int index = parent.getIndex(cell);
        result = index.toString() + CellPath.PATH_SEPARATOR + result;

        cell = parent;
        parent = cell.getParent();
      }
    }

    return (result.length > 1) ? result.substring(0, result.length - 1) : "";
  }

  /**
   * Returns the path for the parent of the cell represented by the given
   * path. Returns null if the given path has no parent.
   * 
   * @param path Path whose parent path should be returned.
   */
  static String getParentPath(String path) {
    if (path != null) {
      int index = path.lastIndexOf(CellPath.PATH_SEPARATOR);

      if (index >= 0) {
        return path.substring(0, index);
      } else if (path.length > 0) {
        return "";
      }
    }

    return null;
  }

  /**
   * Returns the cell for the specified cell path using the given root as the
   * root of the path.
   * 
   * @param root Root cell of the path to be resolved.
   * @param path String that defines the path.
   * @return Returns the cell that is defined by the path.
   */
  static ICell resolve(ICell root, String path) {
    ICell parent = root;
    List<String> tokens = path.split(PATH_SEPARATOR);//Pattern.quote(PATH_SEPARATOR));

    for (int i = 0; i < tokens.length; i++) {
      parent = parent.getChildAt(int.parse(tokens[i]));
    }

    return parent;
  }

  /**
   * Compares the given cell paths and returns -1 if cp1 is smaller, 0 if
   * cp1 is equal and 1 if cp1 is greater than cp2.
   */
  static int compare(String cp1, String cp2) {
//    StringTokenizer p1 = new StringTokenizer(cp1, CellPath.PATH_SEPARATOR);
//    StringTokenizer p2 = new StringTokenizer(cp2, CellPath.PATH_SEPARATOR);
    List<String> l1 = cp1.split(PATH_SEPARATOR);//Pattern.quote(PATH_SEPARATOR));
    List<String> l2 = cp2.split(PATH_SEPARATOR);//Pattern.quote(PATH_SEPARATOR));
    Iterator<String> p1 = l1.iterator;//Pattern.quote(PATH_SEPARATOR));
    Iterator<String> p2 = l2.iterator;//Pattern.quote(PATH_SEPARATOR));
    int comp = 0;

//    while (p1.hasMoreTokens() && p2.hasMoreTokens()) {
//      String t1 = p1.nextToken();
//      String t2 = p2.nextToken();
    while (p1.moveNext() && p2.moveNext()) {
      String t1 = p1.current;
      String t2 = p2.current;

      if (t1 != t2) {
        if (t1.length == 0 || t2.length == 0) {
          comp = t1.compareTo(t2);
        } else {
          comp = int.parse(t1).compareTo(int.parse(t2));
        }

        break;
      }
    }

    // Compares path length if both paths are equal to this point
    if (comp == 0) {
      int t1 = l1.length;
      int t2 = l2.length;

      if (t1 != t2) {
        comp = (t1 > t2) ? 1 : -1;
      }
    }

    return comp;
  }

}
